@@ -1,7 +1,7 @@
 /*
  * Hydrogen
  * Copyright(c) 2002-2008 by Alex >Comix< Cominu [comix@users.sourceforge.net]
- * Copyright(c) 2008-2023 The hydrogen development team [hydrogen-devel@lists.sourceforge.net]
+ * Copyright(c) 2008-2024 The hydrogen development team [hydrogen-devel@lists.sourceforge.net]
  *
  * http://www.hydrogen-music.org
  *
@@ -353,11 +353,6 @@ public:
 	 * Note::__lead_lag times the value calculated by this function.
 	 */
 	long long		getLeadLagInFrames( double fTick );
-	/** Calculates time offset (in frames) #m_pQueuingPosition will be
-	 * ahead of the #m_pTransportPosition.
-	 *
-	 * \return Frame offset*/
-	long long getLookaheadInFrames();
 
 	double getSongSizeInTicks() const;
 
@@ -440,6 +435,8 @@ public:
 	 * Required to not end unit tests prematurely.
 	 */
 	int getEnqueuedNotesNumber() const;
+
+	const QStringList getSupportedAudioDrivers() const;
 	
 	/** Formatted string version for debugging purposes.
 	 * \param sPrefix String prefix which will be added in front of
@@ -468,6 +465,7 @@ public:
 	friend bool CoreActionController::deleteTempoMarker( int );
 	friend bool CoreActionController::locateToTick( long nTick, bool );
 	friend bool CoreActionController::activateSongMode( bool );
+	friend bool CoreActionController::activateLoopMode( bool );
 	/** Is allowed to set m_state to State::Ready via setState()*/
 	friend int FakeDriver::connect();
 	friend void JackAudioDriver::updateTransportPosition();
@@ -511,13 +509,8 @@ private:
 	 * Takes all notes from the currently playing patterns, from the
 	 * MIDI queue #m_midiNoteQueue, and those triggered by the
 	 * metronome and pushes them onto #m_songNoteQueue for playback.
-	 *
-	 * \return
-	 * - 0 - on success
-	 * - -1 - if in Hydrogen is in Song::Mode::Song, looping was
-	 * deactivated, and the end of the song was reached.
 	 */
-	int				updateNoteQueue( unsigned nIntervalLengthInFrames );
+	void			updateNoteQueue( unsigned nIntervalLengthInFrames );
 	void 			processAudio( uint32_t nFrames );
 	long long 		computeTickInterval( double* fTickStart, double* fTickEnd, unsigned nIntervalLengthInFrames );
 	void			updateBpmAndTickSize( std::shared_ptr<TransportPosition> pTransportPosition );
@@ -547,6 +540,7 @@ private:
 	 */
 	void			locateToFrame( const long long nFrame );
 	void			incrementTransportPosition( uint32_t nFrames );
+	bool			isEndOfSongReached( std::shared_ptr<TransportPosition> pPos ) const;
 	void			updateTransportPosition( double fTick, long long nFrame,
 											 std::shared_ptr<TransportPosition> pPos );
 	void			updateSongTransportPosition( double fTick, long long nFrame,
@@ -586,11 +580,17 @@ private:
 	 */
 	void handleDriverChange();
 
+	/** In order to properly support #H2Core::Song::LoopMode::Finishing -
+	 * transport was already looped a couple of times and the user is pressing
+	 * the loop button again to deactivate loop mode - we have to capture the
+	 * number of loops already applied. */
+	void handleLoopModeChanged();
+
 	/**
 	 * Called whenever Hydrogen switches from #Song::Mode::Song into
 	 * #Song::Mode::Pattern or the other way around.
 	 */
-	void switchMode();
+	void handleSongModeChanged();
 
 	Sampler* 			m_pSampler;
 	Synth* 				m_pSynth;
@@ -691,6 +691,24 @@ private:
 	float 			m_fNextBpm;
 	double m_fLastTickEnd;
 	bool m_bLookaheadApplied;
+
+	/**
+	 * Attempts to dynamically load the JACK 2 shared library
+	 * and stores the result in #m_bJackSupported.
+	 */
+	void checkJackSupport();
+
+	/**
+	 * Whether or not the shared library of the JACK server could be
+	 * found on the system at runtime.
+	 */
+	bool m_bJackSupported;
+
+	QStringList m_supportedAudioDrivers;
+
+	/** Indicates how many loops the transport already did when the user presses
+	 * the Loop button again. */
+	int m_nLoopsDone;
 };
 
 
@@ -812,6 +830,9 @@ inline std::shared_ptr<Instrument> AudioEngine::getMetronomeInstrument() const {
 inline int AudioEngine::getEnqueuedNotesNumber() const {
 	return m_songNoteQueue.size();
 }
+inline const QStringList AudioEngine::getSupportedAudioDrivers() const {
+	return m_supportedAudioDrivers;
+}
 };
 
 #endif
